/********************************************************
  BCL Example Problems
  ====================

  file xbdlvriis2rep.c
  ````````````````````
  Transportation problem (infeasible data).
  Repairing infeasibility.
  - Using Optimizer functions -

  (c) 2008-2024 Fair Isaac Corporation
      author: S.Heipcke, Jan. 2008, rev. Mar. 2011
********************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "xprb.h"
#include "xprs.h"

#define NSupp 10                        /* Number of suppliers */
#define NCust 7                         /* Number of customers */
#define MaxArcs 100                     /* Max. num. of non-zero cost values */

#define VANFILE XPRBDATAPATH "/delivery/ifvan.dat"    /* Van data file */
#define COSTFILE XPRBDATAPATH "/delivery/cost.dat"    /* Cost data file */

void printsolution(XPRBprob prob, int scode, XPRBvar x[][NCust]);
void printsolution2(XPRSprob op, int scode, XPRBvar x[][NCust]);

/****DATA****/
/* Supplier:      London  Luton  B'ham Bristl  Derby Stckpt   York */
double SUPPLY[] = {140.0, 200.0,  50.0,  10.0, 400.0, 200.0,  20.0,
/* Supplier: Derby  Soton Scnthp */
             90.0,  30.0,  12.0};
/* Customer:       London Livpol Doncst   York   Hull  Manchr Shffld */
double DEMAND[] = {1230.3, 560.4, 117.1, 592.8, 310.0, 1247.0,  86.0};

double COST[NSupp][NCust];        /* Cost per supplier-customer pair */

double IFVAN[NSupp][NCust];       /* Non-zero if route uses vans instead
                                     of lorries */
double VANCAP=40.0;               /* Capacity on routes that use vans */

/***********************************************************************/

void moddelivery(XPRBprob prob)
{
 XPRBctr ctr, CSupply[NSupp], CDemand[NCust];
 int s,c;
 XPRBvar x[NSupp][NCust];

 int ncol, nrow, scode;
 double *lrp, *grp, *lbp, *ubp;

 XPRSprob op;

/****VARIABLES****/
 for(s=0;s<NSupp;s++)
  for(c=0; c<NCust; c++)
   x[s][c]=XPRBnewvar(prob,XPRB_PL,XPRBnewname("x_s%d_%d",s,c), 0, XPRB_INFINITY);

/****OBJECTIVE****/
 ctr = XPRBnewctr(prob,"OBJ",XPRB_N);
 for(s=0;s<NSupp;s++)             /* Objective: Minimize total cost */
  for(c=0; c<NCust; c++)
   XPRBaddterm(ctr, x[s][c], COST[s][c]);
 XPRBsetobj(prob,ctr);            /* Set objective function */

/****CONSTRAINTS****/
 for(c=0; c<4; c++)               /* Satisfy demand of each customer */
 {
  CDemand[c] = XPRBnewctr(prob,"Demand", XPRB_G);
  for(s=0;s<5;s++) XPRBaddterm(CDemand[c], x[s][c], 1);
  XPRBaddterm(CDemand[c], NULL, DEMAND[c]);
 }

 for(c=4; c<NCust; c++)           /* Satisfy demand of each customer */
 {
  CDemand[c] = XPRBnewctr(prob,"Demand", XPRB_G);
  for(s=5;s<NSupp;s++) XPRBaddterm(CDemand[c], x[s][c], 1);
  XPRBaddterm(CDemand[c], NULL, DEMAND[c]);
 }

 for(s=0;s<5;s++)                 /* Keep within supply at each supplier*/
 {
  CSupply[s] = XPRBnewctr(prob,"Supply",XPRB_L);
  for(c=0; c<4; c++)
   XPRBaddterm(CSupply[s], x[s][c], 1);
  XPRBaddterm(CSupply[s], NULL, SUPPLY[s]);
 }

 for(s=5;s<NSupp;s++)             /* Keep within supply at each supplier*/
 {
  CSupply[s] = XPRBnewctr(prob,"Supply",XPRB_L);
  for(c=4; c<NCust; c++)
   XPRBaddterm(CSupply[s], x[s][c], 1);
  XPRBaddterm(CSupply[s], NULL, SUPPLY[s]);
 }

/****BOUNDS****/
 for(s=0;s<NSupp;s++)
  for(c=0; c<NCust; c++)
   if(IFVAN[s][c]!=0) XPRBsetub(x[s][c], VANCAP);

/****SOLVING + OUTPUT****/
 XPRBsetsense(prob,XPRB_MINIM);   /* Set objective sense to minimization */
/* XPRBexportprob(prob, XPRB_LP, "infeasible"); */
 XPRBlpoptimize(prob,"");         /* Solve the LP-problem */

 printf("LP status: %d\n", XPRBgetlpstat(prob));
 if (XPRBgetlpstat(prob)==XPRB_LP_OPTIMAL)
 {
  printf("Objective: %g\n", XPRBgetobjval(prob));  /* Get objective value */
 }
 else if (XPRBgetlpstat(prob)==XPRB_LP_INFEAS)
 {
  op = XPRBgetXPRSprob(prob);     /* Retrieve the Optimizer problem */

/**** Trying to fix infeasibilities ****/
/*
lrp: (affects = and <= rows)
   ax - aux_var  = b
   ax - aux_var <= b
grp: (affects = and >= rows)
   ax + aux_var  = b
   ax + aux_var >= b
lbp:
   x_i + aux_var >= l
ubp:
   x_i - aux_var <= u
*/

/**** Simplified infeasibility repair:
      specifying preferences per constraint/bound type ****/

  printf("\n**** Repair infeasibility:\n");
  XPRSrepairinfeas(op, &scode, 'c', 'o', ' ', 10, 9, 0, 20, 0.001);
  printsolution2(op, scode, x);         /* Print out the solution values */

/**** Weighted infeasibility repair:
      specifying preferences for every constraint/bound separately ****/

  XPRSgetintattrib(op, XPRS_ORIGINALCOLS, &ncol);
  XPRSgetintattrib(op, XPRS_ORIGINALROWS, &nrow);
  lrp = (double *)malloc(nrow*sizeof(double));
  grp = (double *)malloc(nrow*sizeof(double));
  lbp = (double *)malloc(ncol*sizeof(double));
  ubp = (double *)malloc(ncol*sizeof(double));
  memset(lrp, 0, nrow*sizeof(double));
  memset(grp, 0, nrow*sizeof(double));
  memset(lbp, 0, ncol*sizeof(double));
  memset(ubp, 0, ncol*sizeof(double));

/* Relax bounds due to van capacity */
/* Repairweightedinfeas for upper bounds of concerned flow variables */

  for(s=0; s<NSupp; s++)
   for(c=0; c<NCust; c++)
    if(IFVAN[s][c]!=0) ubp[XPRBgetcolnum(x[s][c])] = 20;

  printf("\n**** Relax van capacity:\n");
  XPRSrepairweightedinfeas(op, &scode, lrp, grp, lbp, ubp, 'd', 0.001, "");
  printsolution2(op, scode, x);

/* Relax supply limits (may buy in additional quantities) */
/* Repairinfeas for 'less or equal' side of Supply constraints */

  for(s=0; s<NSupp; s++) lrp[XPRBgetrownum(CSupply[s])] = 10;

  printf("\n**** Relax supply limits:\n");
  XPRSrepairweightedinfeas(op, &scode, lrp, grp, lbp, ubp, 'd', 0.001, "");
  printsolution(prob, scode, x);

/* Relax demand constraints (may not satisfy all customers) */
/* Repairinfeas for 'greater or equal' side of Demand constraints */

  for(c=0; c<NCust; c++) grp[XPRBgetrownum(CDemand[c])] = 9;

  printf("\n**** Relax demand constraints:\n");
  XPRSrepairweightedinfeas(op, &scode, lrp, grp, lbp, ubp, 'd', 0.001, "");
  printsolution2(op, scode, x);

  free(lrp);
  free(grp);
  free(lbp);
  free(ubp);

 }

}

/***********************************************************************/

    /**** Read data from files ****/
void readdata(void)
{
 FILE *datafile;
 int s,c;

        /* Initialize data tables to 0 */
 for(s=0; s<NSupp; s++)
  for(c=0; c<NCust; c++)
  {
   COST[s][c] = 0;
   IFVAN[s][c] = 0;
  }
        /* Read the demand data file */
 datafile=fopen(COSTFILE,"r");
 for(s=0;s<NSupp;s++)
  XPRBreadarrlinecb(XPRB_FGETS, datafile, 99, "g,", COST[s], NCust);
 fclose(datafile);

        /* Read the van data file */
 datafile=fopen(VANFILE,"r");
 for(s=0;s<NSupp;s++)
  XPRBreadarrlinecb(XPRB_FGETS, datafile, 99, "g,", IFVAN[s], NCust);
 fclose(datafile);
}

    /**** Print out the solution values ****/
void printsolution(XPRBprob prob, int scode, XPRBvar x[][NCust])
{
 int s,c;
 double sup,dem;
 char *rstat[] = {"relaxed optimum found", "relaxed problem infeasible",
  "relaxed problem unbounded", "solution nonoptimal for original objective",
  "error", "numerical instability"};

 printf("Status: %s\n", rstat[scode]);
 if(scode==0)
 {
  XPRBsync(prob, XPRB_XPRS_SOL);
  for(s=0; s<NSupp; s++)
   for(c=0; c<NCust; c++)
    if(XPRBgetsol(x[s][c])>0.01)
     printf("%s:%g ", XPRBgetvarname(x[s][c]), XPRBgetsol(x[s][c]));
  printf("\n");
  printf("Violations:\n");
  for(c=0; c<NCust; c++)
  {
   sup=0;
   for(s=0; s<NSupp; s++) sup+=XPRBgetsol(x[s][c]);
   if(sup<DEMAND[c]) printf(" Customer %d: %g\n", c, DEMAND[c]-sup);
  }
  for(s=0; s<NSupp; s++)
  {
   dem=0;
   for(c=0; c<NCust; c++) dem+=XPRBgetsol(x[s][c]);
   if(dem>SUPPLY[s]) printf(" Supplier %d: %g\n", s, dem-SUPPLY[s]);
  }
  for(s=0; s<NSupp; s++)
   for(c=0; c<NCust; c++)
    if(IFVAN[s][c]!=0 && VANCAP<XPRBgetsol(x[s][c]))
     printf(" Van %d-%d: %g\n", s, c, XPRBgetsol(x[s][c])-VANCAP);
 }
}

void printsolution2(XPRSprob op, int scode, XPRBvar x[][NCust])
{
 int s,c,ncol;
 double sup,dem;
 char *rstat[] = {"relaxed optimum found", "relaxed problem infeasible",
  "relaxed problem unbounded", "solution nonoptimal for original objective",
  "error", "numerical instability"};
 double *sol;

 printf("Status: %s\n", rstat[scode]);
 if(scode==0)
 {
  XPRSgetintattrib(op, XPRS_ORIGINALCOLS, &ncol);
  sol = (double *)malloc(ncol * sizeof(double));
  XPRSgetlpsol(op, sol, NULL, NULL, NULL);     /* Get the solution values */

  for(s=0; s<NSupp; s++)
   for(c=0; c<NCust; c++)
    if(XPRBgetcolnum(x[s][c])>=0 && sol[XPRBgetcolnum(x[s][c])]>0.01)
     printf("%s:%g ", XPRBgetvarname(x[s][c]), sol[XPRBgetcolnum(x[s][c])]);
  printf("\n");
  printf("Violations:\n");
  for(c=0; c<NCust; c++)
  {
   sup=0;
   for(s=0; s<NSupp; s++)
    if(XPRBgetcolnum(x[s][c])>=0) sup+=sol[XPRBgetcolnum(x[s][c])];
   if(sup<DEMAND[c]) printf("  Customer %d: %g\n", c, DEMAND[c]-sup);
  }
  for(s=0; s<NSupp; s++)
  {
   dem=0;
   for(c=0; c<NCust; c++)
    if(XPRBgetcolnum(x[s][c])>=0) dem+=sol[XPRBgetcolnum(x[s][c])];
   if(dem>SUPPLY[s]) printf("  Supplier %d: %g\n", s, dem-SUPPLY[s]);
  }
  for(s=0; s<NSupp; s++)
   for(c=0; c<NCust; c++)
    if(IFVAN[s][c]!=0 && XPRBgetcolnum(x[s][c])>=0 &&
       VANCAP<sol[XPRBgetcolnum(x[s][c])])
     printf("  Van %d-%d: %g\n", s, c, sol[XPRBgetcolnum(x[s][c])]-VANCAP);
 }
}


/***********************************************************************/

int main(int argc, char **argv)
{
 XPRBprob prob;

 prob=XPRBnewprob("Delivery");   /* Initialize a new problem in BCL */
 readdata();                     /* Data input from file */
 moddelivery(prob);              /* Problem formulation & solving */

 return 0;
}


